unit cxdconsts;

interface
{$Z1}
{$WARNINGS OFF}
type
  TContestant = (ctDelphi, ctFastString, ctAJSearch, ctMWSearch, ctcxpos);
  TContestants = set of TContestant;

  TMethode = (mtPos, mtCount, mtReplace, mtPosFile, mtCountFile, mtReplaceFile);
  TMethodes = set of TMethode;

  TContestantMethodes = packed array[TContestant] of TMethodes;

  TReturn = (rtResult, rtTimeTotal, rtTimeAverage, rtPerfRatio, rtRank);
  TReturns = set of TReturn;
  TProductItem = (piReturn, piContestant, piMethode); // Dimension, 3D

  // TContestant (5 members) fit into 3 bits (since 2^3=8, can handle max 8 members)
  // TMethode (6 members) and TReturn (5 members) all also fit into 3 bits

  // we must use the widest type bits for EncodeShift, here at least 3 bits
  // no problem if it is wider, using 8bits (max 256 members) is even more
  // convenient, because that would be human readable

  // total bitmask used  = (TProduct item members * EncodeShift) = 9 bits
  // total bitmask should not exceed integer size (32-bits), actually it could,
  // if you change integer to int64 which will double the capacity, just slower

  // See also encoding/decoding and EncodeShift below

  TProducts = packed array[TContestant] of packed array of TMethode;
  TProductReturns = packed array[TReturn] of packed array of TProducts;

const
  ContestantsName: packed array[TContestant] of string =
  ('Delphi', 'FastStrings', 'AJSearch', 'MwSearch', 'cxpos');
  MethodesName: packed array[TMethode] of string =
  ('Locate (Pos)', 'Count', 'Replace', 'Pos in File', 'Count in File', 'Replace in File');
  ReturnsCaption: packed array[TReturn] of string =
  ('Function Result', 'TOTAL Time (ns)', 'Average time (ns)', 'Performance', 'Rank');
  ContestantsDescription: packed array[TContestant] of string = (
    'Delphi standard Pos function',
    'FastStrings & FastStringFuncs'^m'Version: 3.2'^m'by: Peter Morris',
    'AJSearch TSearch/TFileSearch component'^m'Version: 2.2'^m'by: Angus Johnson',
    'MWSearch, Version: 2.2'^m'by: Martin.Waldenburg',
    'cxpos unit, Version: 2.0.3.2'^m'by: aa & Adrian Hafizh'
    );

const
  DefaultEnabledContestants = //[ctDelphi, ctFastString, ctcxpos];
  //[low(TContestant)..high(TContestant)];
  [low(TContestant)..high(TContestant)] - [ctMWSearch];
  //[ctAJSearch];
  DefaultChecked_1 = [low(TMethode)..pred(high(TMethode))];
  DefaultChecked_2 = DefaultChecked_1 - [mtReplace];

  DefaultCheckedDelphi = [mtPos]; //, mtCount];
  DefaultCheckedFastString = DefaultChecked_1;//DefaultCheckedDelphi; // + [mtReplace];
  DefaultCheckedAJSearch = DefaultChecked_2; //DefaultCheckedFastString;
  DefaultCheckedMWSearch = DefaultChecked_2; //DefaultCheckedFastString;
  DefaultCheckedcxpos = DefaultChecked_1;

  DefaultCheckedMethodes: TContestantMethodes = (DefaultCheckedDelphi,
    DefaultCheckedFastString, DefaultCheckedAJSearch,
    DefaultCheckedMWSearch, DefaultCheckedcxpos
    );

  DefaultCheckedMethode = ([]);

  ContestantCount = ord(high(TContestant)) + 1;
  MethodeCount = ord(high(TMethode)) + 1;
  ReturnCount = ord(high(TReturn)) + 1;
  ContestantProduct = ContestantCount * MethodeCount;
  ProductItemsCount: array[TProductItem] of integer =
  (ReturnCount, ContestantCount, MethodeCount);

  YES = TRUE;
  NAY = not TRUE; // = none is better
  NOO = not TRUE; // = strongly discouraged                            ordnums
  OH = boolean(nil); // = up to you
  OHNOO = not not not YES; // = don't you even dare to try
  // so you can say... SysAdminMode.Disabled := OH-NOO

const
  //we avoid using inifiles
  CONFIG_FILE: string = '';
  CONFIG_HEADER = '[cxdemo-settings]';

type
//hard-coded, MUST be synchronized with type tsettings, sorry :))
//type
//  TSettings = packed record
  TConfigIndex = (
    ciHeader,
    ciSaveConfig,
    ciSubStrText, ciReplacementText, ciFileNameText,
    ciSubStrList, ciReplacementList, ciFileNameList,
    ciIgnoreCase, ciKeepWorkFiles,
    ciStartPos, ciLoopCount, ciLocation,
    ciEnabledContestants, ciCheckedMethodes
    );

const
  ConfigIndexCount = ord(high(TConfigIndex)) + 1;

type
  TSettings = packed record
    SubStrText, SubStrList: string;
    ReplacementText, ReplacementList: string;
    FileNameText, FileNameList: string;
    SaveConfig: string;
    IgnoreCase, KeepWorkFiles: boolean;
    StartPos, LoopCount, Location: integer;
    EnabledContestants: TContestants;
    CheckedMethodes: TContestantMethodes;
  end;

const
  _ciHeader = ciHeader;
  _ciSaveConfig = (ord(ciSaveConfig)); //4;
  _ciSubStrText = (ord(ciSubStrText)); //1;
  _ciReplacementText = (ord(ciReplacementText)); //2;
  _ciFileNameText = (ord(ciFileNameText)); //3;
  _ciSubStrList = (ord(ciSubStrList)); //1;
  _ciReplacementList = (ord(ciReplacementList)); //2;
  _ciFileNameList = (ord(ciFileNameList)); //3;
  _ciIgnoreCase = (ord(ciIgnoreCase)); //5;
  _ciKeepWorkFiles = (ord(ciKeepWorkFiles)); //5+;
  _ciStartPos = (ord(ciStartPos)); //6;
  _ciLoopCount = (ord(ciLoopCount)); //7;
  _ciLocation = (ord(ciLocation)); //8;
  _ciEnabledContestants = (ord(ciEnabledContestants)); //9;
  _ciCheckedMethodes = (ord(ciCheckedMethodes)); //10;

var
  GlobalSettings: TSettings = (
    SaveConfig: '-1';
    LoopCount: 1;
    //KeepWorkFile: FALSE;
    EnabledContestants: DefaultEnabledContestants;
    CheckedMethodes: (DefaultCheckedDelphi,
    DefaultCheckedFastString, DefaultCheckedAJSearch,
    DefaultCheckedMWSearch, DefaultCheckedcxpos);
    );
  //

//function HexStr(const b: byte): string;
//function HexBuf(const Buffer: pointer; const BufferLength: integer; const Delimiter: Char): string;

function EncodeProducts(const Return: TReturn; const Contestant: TContestant; const methode: TMethode): integer;

function DecodeReturn(Value: integer): TReturn;
function DecodeContestant(Value: integer): TContestant;
function DecodeMethode(Value: integer): TMethode;

//~ function Hexs(const byte: byte; const uppercase: boolean = YES): string; overload;
//~ function Hexs(const word: word; const uppercase: boolean = YES): string; overload;
//~ function Hexs(const integer: integer; const uppercase: boolean = YES): string; overload;
//~ function Hexs(const I: int64; const uppercase: boolean = YES): string; overload;
//~ function Hexs(const Buffer: pointer; const BufferLength: integer;
//~   const Delimiter: Char = #0; const Uppercase: boolean = YES): string; overload;
//~ function Hexs(const Buffer: pointer; const BufferLength: integer;
//~   const Uppercase: boolean; const Delimiter: Char = #0): string; overload;
//~
function MaxOf(A, b: integer): integer; overload;
function MinOf(a, B: integer): integer; overload;
function intoStr(const I: integer): string;

function Blocks(const S: string; const BlockLen: integer = 3;
  const delimiter: string = ' '; const LeftWise: Boolean = FALSE): string; overload;
function Blocks(const I: integer; const BlockLen: integer = 3;
  const delimiter: string = ' '; const LeftWise: Boolean = FALSE): string; overload;
function Blocks(const S: string; const Delimiter: string = ' '; BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE): string; overload;
function Blocks(const I: integer; const Delimiter: string = ' '; BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE): string; overload;


implementation
//uses OrdNums;

const
  ByteBits = 8;
  CustomBits = 4;
  EncodeShift = CustomBits; //ByteBits;
  EncodeMask = 1 shl EncodeShift - 1;

function EncodeProducts(const Return: TReturn; const Contestant: TContestant; const methode: TMethode): integer;
// here we permit EncodeShift <> ByteBits
const
  rk = EncodeMask shl (ord(piReturn) * EncodeShift);
  ck = EncodeMask shl (ord(piContestant) * EncodeShift);
  mk = EncodeMask shl (ord(piMethode) * EncodeShift);
var
  r, c, m: integer;
begin
  r := rk and (ord(Return) shl (ord(piReturn) * EncodeShift));
  c := ck and (ord(Contestant) shl (ord(piContestant) * EncodeShift));
  m := mk and (ord(methode) shl (ord(piMethode) * EncodeShift));

  Result := r or c or m;
end;

function DecodeProduct(const Value: integer; const ProductItem: TProductItem): byte;
begin
  Result := (Value shr (EncodeShift * ord(ProductItem))) and EncodeMask;
  if Result >= ProductItemsCount[ProductItem] then
    Result := pred(ProductItemsCount[ProductItem])
end;

function DecodeReturn(Value: integer): TReturn;
begin
  Result := TReturn(DecodeProduct(Value, piReturn));
end;

function DecodeContestant(Value: integer): TContestant;
begin
  Result := TContestant(DecodeProduct(Value, piContestant));
end;

function DecodeMethode(Value: integer): TMethode;
begin
  Result := TMethode(DecodeProduct(Value, piMethode));
end;

function MaxOf(A, b: integer): integer; asm
   cmp a, b; jge @end
  mov a, b
  @end:
end;

function MinOf(a, B: integer): integer; asm
  cmp a, b; jle @end
  mov a, b
  @end:
end;

//~ // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~ // Note:
//~ // all of the helper functions below are excerpted
//~ // from Acommon & MinMaxMid unit, by the same authors
//~ // those are subject to change, please get a more
//~ // recent copy of them
//~ // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~ // OBSOLETE! still containing bugs, use OrdNums unit instead
//~
function intoStr(const I: integer): string;
begin
  Str(i, Result);
end;
//~
//~ const
//~   HEXNUM_UPPERCASE: string = '0123456789ABCDEF';
//~   HEXNUM_LOWERCASE: string = '0123456789abcdef';
//~
//~ function Hexs(const byte: byte; const uppercase: boolean = YES): string;
//~ asm
//~   push ebx
//~   mov ah, uppercase; push eax
//~
//~   mov eax, Result           // where the result will be stored
//~   mov edx, 2                // how much length of str requested
//~   call System.@LStrSetLength// result: new allocation pointer in EAX
//~   mov edx, [eax]            // eax contains the new allocated pointer
//~                             // we got the storage as well at once
//~   mov ebx, HEXNUM_UPPERCASE // init ebx with default value
//~   pop eax                   // ah = uppercase flag
//~   test ah, 1                // is uppercase flag = TRUE?
//~     jne @skipsetlocase      // nz = yes, then dont bother to kowercase
//~   mov ebx, HEXNUM_LOWERCASE // uppercase flags = FALSE
//~
//~   @skipsetlocase:
//~   mov ah, al                // save the value first
//~
//~   and al, $f; xlat          // strip high nibbles, translate
//~   mov edx+1, al             // save to allocated string / last position
//~
//~   mov al, ah
//~   shr eax, 3 * 4             // shift 3 nibbles away (high nibble of AH)
//~   and al, $f; xlat          // strip high nibbles, translate
//~   mov [edx], al             // save to allocated string / first position
//~
//~   pop ebx
//~ end;
//~
//~ function Hexs(const word: word; const uppercase: boolean = YES): string; overload;
//~ const WordDigits = 4;
//~ asm
//~   push ebx
//~   xchg ah, al; push ax; shl eax, 16; pop ax
//~   mov al, uppercase; push eax
//~
//~   mov eax, Result           // where the result will be stored
//~   mov edx, WordDigits       // how much length of str requested
//~   call System.@LStrSetLength// result: new allocation pointer in EAX
//~   mov edx, [eax]            // eax contains the new allocated pointer
//~                             // we got the storage as well at once
//~   mov ebx, HEXNUM_UPPERCASE // init ebx with default value
//~   pop eax                   // al = uppercase flag
//~   test al, 1                // is uppercase flag = TRUE?
//~     jne @skipsetlocase      // nz = yes, then dont bother to kowercase
//~   mov ebx, HEXNUM_LOWERCASE // uppercase flags = FALSE
//~
//~   @skipsetlocase:
//~
//~   mov al, ah                // get the high byte
//~   and al, $f; xlat          // strip high nibbles, translate
//~   mov edx+3, al             // save to allocated string / last position
//~
//~   mov al, ah; shr al, 4     // reget the high byte; get high nibble
//~   and al, $f; xlat          // strip high nibbles, translate
//~   mov edx+2, al             // save to allocated string / first position
//~
//~   shr eax, 16               //
//~   mov ah, al                // we dont again need ah
//~
//~   and al, $f; xlat
//~   mov edx+1, al
//~
//~   mov al, ah
//~   shr al, 4
//~   and al, $f; xlat
//~   mov [edx], al
//~
//~   pop ebx
//~ end;
//~
//~ function Hexs(const integer: integer; const uppercase: boolean = YES): string; overload;
//~ const
//~   IntDigits = 8;
//~   bswap_eax = $C80F;
//~ begin
//~   asm
//~     push ebx
//~     db $f,$c8                 // bswap eax; $c9=ecx $ca=edx $cb=ebx
//~     mov integer, eax
//~
//~     mov eax, Result           // where the result will be stored
//~     mov edx, IntDigits        // how much length of str requested
//~     call System.@LStrSetLength// result: new allocation pointer in EAX
//~     mov edx, [eax]            // eax contains the new allocated pointer
//~                               // we got the storage as well at once
//~     mov ebx, HEXNUM_UPPERCASE // init ebx with default value
//~
//~     test uppercase, 1;        // is uppercase flag = TRUE?
//~       jne @skipsetlocase      // nz = yes, then dont bother to kowercase
//~     mov ebx, HEXNUM_LOWERCASE // uppercase flags = FALSE
//~     @skipsetlocase:
//~
//~     mov eax, integer
//~     mov ecx, IntDigits / 2
//~     mov ch, al                // initialize
//~
//~   @Loop:
//~     shr al, 4
//~     and al, $f; xlat
//~     mov [edx], al
//~     lea edx, edx +1
//~
//~     mov al, ch
//~     and al, $f; xlat
//~     mov [edx], al
//~     lea edx, edx +1
//~
//~     shr eax, 8
//~     mov ch, al
//~     dec cl
//~     jnz @Loop
//~
//~     pop ebx
//~   end;
//~ end;
//~
//~ function Hexs(const I: int64; const uppercase: boolean = YES): string; overload;
//~ const I64digits = 16;
//~ type
//~   i64 = packed record
//~     lo, hi: integer
//~   end;
//~
//~ begin
//~   asm
//~     push ebx
//~     mov eax, I.i64.lo
//~     db $f,$c8             // bswap eax
//~     mov I.i64.lo, eax
//~
//~     mov eax, I.i64.hi
//~     db $f,$c8             // bswap eax
//~     mov I.i64.hi, eax
//~
//~
//~     mov eax, Result           // where the result will be stored
//~     mov edx, 16                // how much length of str requested
//~     call System.@LStrSetLength// result: new allocation pointer in EAX
//~     mov edx, [eax]            // eax contains the new allocated pointer
//~     mov ebx, HEXNUM_UPPERCASE // init ebx with default value
//~     test uppercase, 1;        // is uppercase flag = TRUE?
//~       jne @skipsetlocase      // nz = yes, then dont bother to kowercase
//~     mov ebx, HEXNUM_LOWERCASE // uppercase flags = FALSE
//~
//~     @skipsetlocase:
//~
//~     mov eax, I.i64.hi
//~     mov ecx, 4; mov ch, al
//~
//~   @Loop1:
//~     shr al, 4
//~     and al, $f; xlat
//~     mov [edx], al
//~     lea edx, edx +1
//~
//~     mov al, ch; //shr al, 4
//~     and al, $f; xlat
//~     mov [edx], al
//~     lea edx, edx +1
//~
//~     shr eax, 8
//~     mov ch, al
//~     dec cl
//~     jnz @Loop1
//~
//~     mov eax, I.i64.lo
//~     mov ecx, 4; mov ch, al
//~
//~   @Loop2:
//~     shr al, 4
//~     and al, $f; xlat
//~     mov [edx], al
//~     lea edx, edx +1
//~
//~     mov al, ch; //shr al, 4
//~     and al, $f; xlat
//~     mov [edx], al
//~     lea edx, edx +1
//~
//~     shr eax, 8
//~     mov ch, al
//~     dec cl
//~     jnz @Loop2
//~
//~     pop ebx
//~   end;
//~ end;
//~
//~
//~ function Hexs(const Buffer: pointer; const BufferLength: integer;
//~   const Delimiter: Char = #0; const Uppercase: boolean = YES): string;
//~ begin
//~   asm
//~   @@Start:
//~     or eax, eax; jz @@Stop     // insanity checks
//~     or edx, edx; jle @@Stop
//~
//~     push esi; push edi; push ebx
//~     mov esi, buffer
//~     mov edi, bufferlength      // save buflength first!
//~
//~     shl edx, 1                 // edi = Bufferlength * 2
//~     and ecx, $ff               // note: ecx IS delimiter
//~     push ecx                   // save it, will be destroyed by LStrSetLength
//~
//~     cmp ecx, 0; jz @nolimit    // if delim = #0 then skip increase length
//~     lea edx, edx+edi           // ~> inc(edx, edi)
//~     dec edx                    // we don't need trailing delimiter
//~
//~   @nolimit:
//~     mov eax, @Result            // where the result will be stored
//~     call System.@LStrSetLength // result: new allocation pointer in EAX
//~                                // ecx, edx destroyed
//~                                // {var str: AnsiString; newLength: Integer};
//~
//~     mov ecx, edi               // get (original) bufferlength back
//~
//~     //mov edi, eax               // WRONG! eax contains the new allocated pointer
//~                                // we got the storage as well at once
//~     mov edi, [eax]             // eax contains the new allocated pointer
//~
//~     pop edx                    // get delimiter back
//~
//~     mov ebx, HEXNUM_UPPERCASE  // get Translation Table
//~     test Uppercase, 1; jne @skipsetlocase  // is uppercase flag = TRUE?
//~
//~     mov ebx, HEXNUM_LOWERCASE // uppercase flags = FALSE
//~
//~   @skipsetlocase:
//~
//~     cmp edx, 0
//~     Jz @@WithoutDelimiter
//~
//~     dec ecx
//~     jz @LastByte
//~
//~     @@WithDelimiter:
//~       lodsb                    // load byte to AL
//~       mov ah, al               // copy to AH for second nibble translation
//~
//~       shr al, 4                // extract high nibble
//~       xlat; mov [edi], al      // translate and store result
//~       mov al, ah; and al, $f   // extract low nibble; validate
//~       xlat; mov edi+1, al      // translate and store result
//~       mov edi+2, dl            // put delimiter
//~       lea edi, edi+3           // inc edi by 3
//~       dec ecx; jg @@WithDelimiter
//~
//~    @LastByte:
//~       lodsb                    // load byte to AL
//~       mov ah, al               // copy to AH for second nibble translation
//~
//~       shr al, 4                // extract high nibble
//~       xlat; mov [edi], al      // translate and store result
//~       mov al, ah; and al, $f   // extract low nibble; validate
//~       xlat; mov edi+1, al      // translate and store result
//~       //we donot need these anymore...
//~       //lea edi, edi+3           // inc edi by 3
//~       //mov edi+2, dl            // put delimiter
//~       //dec ecx; jg @@WithDelimiter
//~       lea edi, edi +2
//~       jmp @@Done
//~
//~     @@WithoutDelimiter:
//~       lodsb                    // load byte to AL
//~       mov ah, al               // copy to AH for second nibble translation
//~
//~       shr al, 4                // extract high nibble
//~       xlat; mov [edi], al      // translate and store result
//~       mov al, ah; and al, $f   // extract low nibble; validate
//~       xlat; mov edi+1, al      // translate and store result
//~       //mov edi+2, dl            // put delimiter
//~       //lea edi, edi+3           // inc edi by 3
//~       lea edi, edi+2           // inc edi only 2 in this block
//~       dec ecx; jg @@WithoutDelimiter
//~       jmp @@Done
//~
//~     @@Done:
//~     mov eax, @Result
//~     pop ebx; pop edi; pop esi
//~   @@Stop:
//~   end;
//~ end;
//~
//~ function Hexs(const Buffer: pointer; const BufferLength: integer;
//~   const Uppercase: boolean; const Delimiter: Char = #0): string; overload;
//~ begin
//~   Result := Hexs(Buffer, BufferLength, Delimiter, Uppercase);
//~ end;

function Blocks(const S: string; const BlockLen: integer = 3;
  const delimiter: string = ' '; const LeftWise: Boolean = FALSE): string;
const
  MINBLOCKLEN = 2; MAXBLOCKLEN = high(word);
var
  i, x: integer;
begin
  Result := S;
  x := MaxOf(word(BlockLen), MINBLOCKLEN); // max.65535
  i := length(Result) div x;
  if length(Result) mod x = 0 then dec(i);
  if LeftWise then
    for i := i downto 1 do
      insert(delimiter, Result, (i * x) + 1)
  else
    for i := i downto 1 do
      insert(delimiter, Result, length(Result) - (i * x) + 1)
end;

function Blocks(const I: integer; const BlockLen: integer = 3;
  const delimiter: string = ' '; const LeftWise: Boolean = FALSE): string;
begin
  Result := blocks(intoStr(I), BlockLen, delimiter, LeftWise);
  //Hexs(@i, 2);
end;

function Blocks(const S: string; const Delimiter: string = ' '; BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE): string;
begin
  Result := blocks(S, BlockLen, delimiter, LeftWise);
end;

function Blocks(const I: integer; const Delimiter: string = ' '; BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE): string;
begin
  Result := blocks(intoStr(I), BlockLen, delimiter, LeftWise);
end;


end.

